@@ -805,36 +805,6 @@ fhandler_socket_unix::recv_peer_info ()
   return ret;
 }
 
-NTSTATUS
-fhandler_socket_unix::npfs_handle (HANDLE &nph)
-{
-  static NO_COPY SRWLOCK npfs_lock;
-  static NO_COPY HANDLE npfs_dirh;
-
-  NTSTATUS status = STATUS_SUCCESS;
-  OBJECT_ATTRIBUTES attr;
-  IO_STATUS_BLOCK io;
-
-  /* Lockless after first call. */
-  if (npfs_dirh)
-    {
-      nph = npfs_dirh;
-      return STATUS_SUCCESS;
-    }
-  AcquireSRWLockExclusive (&npfs_lock);
-  if (!npfs_dirh)
-    {
-      InitializeObjectAttributes (&attr, &ro_u_npfs, 0, NULL, NULL);
-      status = NtOpenFile (&npfs_dirh, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
-			   &attr, &io, FILE_SHARE_READ | FILE_SHARE_WRITE,
-			   0);
-    }
-  ReleaseSRWLockExclusive (&npfs_lock);
-  if (NT_SUCCESS (status))
-    nph = npfs_dirh;
-  return status;
-}
-
 HANDLE
 fhandler_socket_unix::create_pipe (bool single_instance)
 {
@@ -1696,6 +1666,13 @@ fhandler_socket_unix::connect (const struct sockaddr *name, int namelen)
       conn_unlock ();
       return -1;
     }
+  if (name->sa_family == AF_UNSPEC && get_socket_type () == SOCK_DGRAM)
+    {
+      connect_state (unconnected);
+      peer_sun_path (NULL);
+      conn_unlock ();
+      return 0;
+    }
   connect_state (connect_pending);
   conn_unlock ();
   /* Check validity of name */
